package tools.descartes.prisma.core;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.eclipse.emf.ecore.resource.Resource;

import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPointDescriptions;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationpointsFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.core.exceptions.ContainerConflictException;
import tools.descartes.prisma.model.agent.Agent;
import tools.descartes.prisma.model.agent.AgentFactory;
import tools.descartes.prisma.model.agent.AgentRepository;
import tools.descartes.prisma.model.agent.ContainerFilter;
import tools.descartes.prisma.model.sensor.SensorFactory;
import tools.descartes.prisma.model.sensor.SensorRepository;

public abstract class ContainerScope extends Scope {

	private static final String CONTAINER_DIRECTORY = "container";
	private static final String AGENTS_SUFFIX = ".agents";
	private static final String ADAPTATIONPOINTS_SUFFIX = ".adaptationpoints";
	private static final String SENSORS_SUFFIX = ".sensors";
	private static final String CONTAINER_SUFFIX = ".container";
	private static final String CONTAINER_PROPERTIES_SUFFIX = ".properties";

	private final String type;
	private AgentRepository agents;
	private SensorRepository sensors;
	private AdaptationPointDescriptions adaptationPoints;
	private final Map<String, Container> containers = new HashMap<>();
	private final SystemGlobalScope systemScope;
	
	public ContainerScope(String name, String type, SystemGlobalScope systemScope, ModelRepository repository) {
		super(name, repository);
		this.type = type;
		this.systemScope = systemScope;
	}
	
	@Override
	protected String getBasePath() {
		return type;
	}

	public SystemGlobalScope getSystemScope() {
		return systemScope;
	}

	public SensorRepository getSensors() {
		if (sensors == null) {
			sensors = (SensorRepository) getModelRepository().getResource(getSensorsPath()).getContents().get(0);
		}
		return sensors;
	}

	private String getSensorsPath() {
		return type + "/" + getName() + "/" + getName() + SENSORS_SUFFIX;
	}
	
	public AdaptationPointDescriptions getAdaptationPoints() {
		if (adaptationPoints == null) {
			adaptationPoints = (AdaptationPointDescriptions) getModelRepository().getResource(getAdaptationPointsPath()).getContents().get(0);
		}
		return adaptationPoints;
	}
	
	private String getAdaptationPointsPath() {
		return type + "/" + getName() + "/" + getName() + ADAPTATIONPOINTS_SUFFIX;
	}
	
	public AgentRepository getAgents() {
		if (agents == null) {
			agents = (AgentRepository) getModelRepository().getResource(getAgentsPath()).getContents().get(0);
		}
		return agents;
	}

	private String getAgentsPath() {
		return type + "/" + getName() + "/" + getName() + AGENTS_SUFFIX;
	}
	
	public Container getContainer(String id) {
		Container curContainer = containers.get(id);
		if (curContainer == null) {
			curContainer = (Container) getModelRepository().getResource(getContainerPath(id));
			if (curContainer != null) {
				containers.put(id, curContainer);
			}
		}
		return curContainer;
	}

	private String getContainerPath(String id) {
		return getContainerDirectory() + "/" + id + CONTAINER_SUFFIX;
	}

	private String getContainerPropertiesPath(String id) {
		return getContainerDirectory() + "/" + id + CONTAINER_PROPERTIES_SUFFIX;
	}

	private String getContainerDirectory() {
		return type + "/" + getName() + "/" + CONTAINER_DIRECTORY;
	}

	@Override
	protected void createResources(Transaction transaction) {
		Resource adaptationPointsResource = getModelRepository().getOrCreateResource(transaction,
				getAdaptationPointsPath());
		if (adaptationPointsResource.getContents().isEmpty()) {
			AdaptationPointDescriptions adaptationPoints = AdaptationpointsFactory.eINSTANCE
					.createAdaptationPointDescriptions();
			adaptationPoints.setName(getName());
			adaptationPointsResource.getContents().add(adaptationPoints);
		}
		
		Resource agentsResource = getModelRepository().getOrCreateResource(transaction, getAgentsPath());
		if (agentsResource.getContents().isEmpty()) {
			AgentRepository agents = AgentFactory.eINSTANCE.createAgentRepository();
			agentsResource.getContents().add(agents);
		}
		
		Resource sensorsResource = getModelRepository().getOrCreateResource(transaction, getSensorsPath());
		if (sensorsResource.getContents().isEmpty()) {
			SensorRepository sensors = SensorFactory.eINSTANCE.createSensorRepository();
			sensorsResource.getContents().add(sensors);
		}

		getModelRepository().createFolders(transaction, getContainerDirectory());
	}

	public Agent getAgent(String host, int port) {
		AgentRepository agents = getAgents();
		for (Agent curAgent : agents.getAgents()) {
			if (curAgent.getHost().equals(host) && (curAgent.getPort() == port)) {
				return curAgent;
			}
		}
		return null;
	}

	public Agent addAgent(String host, int port) {
		Agent agent = getAgent(host, port);
		if (agent == null) {
			try (Transaction transaction = getModelRepository().createTransaction()) {
				// 1. Save this agent in the repository
				AgentRepository agents = transaction.attach(getAgents());
				agents.cdoWriteLock().lock();
				agent = AgentFactory.eINSTANCE.createAgent();
				agent.setHost(host);
				agent.setPort(port);
				agents.getAgents().add(agent);
				transaction.commit();

				// 2. See which of the existing containers are observed by this
				// agent
				registerExistingContainersWithAgent(agent);
				
				return transaction.detach(agent);
			} catch (CommitException e) {
				e.printStackTrace();
			}
		}
		return agent;
	}

	public void updateAgent(Agent agent) {
		Set<Container> containersToRemove = new HashSet<>(agent.getContainers());
		for (String name : getModelRepository().listResources(getContainerDirectory())) {
			try {
				Container curContainer = getContainer(name);
				Map<String, String> properties = getModelRepository()
						.loadPropertiesFile(getContainerPropertiesPath(curContainer.getName()));
				if (matchesContainer(agent, properties)) {
					if (containersToRemove.contains(curContainer)) {
						containersToRemove.remove(curContainer);
					} else {
						registerContainerWithAgent(agent, curContainer);
					}
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
		
		if (!containersToRemove.isEmpty()) {
			try (Transaction transaction = getModelRepository().createTransaction()) {
				Agent agentInTransaction = transaction.attach(agent);
				agentInTransaction.cdoWriteLock().lock();
				for (Container curContainer : containersToRemove) {
					agentInTransaction.getContainers().remove(curContainer);
				}
				transaction.commit();
			} catch (CommitException e) {
				e.printStackTrace();
			}
		}
	}

	private void registerExistingContainersWithAgent(Agent agent) {

		for (String name : getModelRepository().listResources(getContainerDirectory())) {
			try {
				Container curContainer = getContainer(name);
				Map<String, String> properties = getModelRepository().loadPropertiesFile(curContainer.getName());
				if (matchesContainer(agent, properties)) {
					registerContainerWithAgent(agent, curContainer);
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}

	public void addContainer(Container container, Map<String, String> properties) {
		String scopeName = properties.get(ContainerPropertyNames.CONTAINER_SCOPE);
		ContainerScope scope = (ContainerScope) getModelRepository().getScope(scopeName);
		if (scope != this) {
			// This container is in a different scope than ours
			scope.waitForCreation();
			scope.addContainer(container, properties);
		} else {
			
			persistContainer(container, properties);
			
			for (Agent curAgent : getAgents().getAgents()) {
				if (matchesContainer(curAgent, properties)) {
					registerContainerWithAgent(curAgent, container);
				}
			}
		}
	}

	private void persistContainer(Container container, Map<String, String> properties) {
		try (Transaction transaction = getModelRepository().createTransaction()) {
			Resource res = getModelRepository().getOrCreateResource(transaction, getContainerPath(container.getName()));
			if (res.getContents().isEmpty()) {
				res.getContents().add(container);
			} else {
				if (!container.equals(res.getContents().get(0))) {
					transaction.rollback();
					throw new ContainerConflictException(container);
				}
			}

			getModelRepository().storePropertiesFile(transaction, getContainerPropertiesPath(container.getName()),
					properties);
			
			transaction.commit();
		} catch (CommitException | IOException e) {
			e.printStackTrace();
		}
	}

	private void registerContainerWithAgent(Agent agent, Container container) {
		try (Transaction transaction = getModelRepository().createTransaction()) {
			Agent agentInTransaction = transaction.attach(agent);
			agentInTransaction.cdoWriteLock().lock();
			if (!agentInTransaction.getContainers().contains(container)) {
				agentInTransaction.getContainers().add(container);
			}
			transaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	private boolean matchesContainer(Agent agent, Map<String, String> properties) {
		for (ContainerFilter curExclude : agent.getExcludeFilters()) {
			String value = properties.get(curExclude.getProperty());
			if (value != null) {
				if (Pattern.matches(curExclude.getFilter(), value)) {
					return false;
				}
			}
		}
		for (ContainerFilter curInclude : agent.getIncludeFilters()) {
			String value = properties.get(curInclude.getProperty());
			if (value != null) {
				if (Pattern.matches(curInclude.getFilter(), value)) {
					return true;
				}
			}
		}
		return false;
	}
}
