package tools.descartes.prisma.core.agent;

import java.io.IOException;
import java.net.InetAddress;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;

import org.eclipse.emf.cdo.CDOAdapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;

import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import tools.descartes.prisma.core.ContainerScope;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.RepositoryConnection;
import tools.descartes.prisma.core.SystemGlobalScope;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.model.agent.Agent;
import tools.descartes.prisma.model.agent.AgentFactory;
import tools.descartes.prisma.model.agent.AgentPackage;
import tools.descartes.prisma.model.agent.AgentRepository;
import tools.descartes.prisma.model.agent.ContainerFilter;

public class AgentController {
	
	private class ContainerListAdapter implements CDOAdapter {

		private Agent target;

		@Override
		public void notifyChanged(Notification msg) {
			if (msg.getFeatureID(Agent.class) == AgentPackage.AGENT__CONTAINERS) {
				if (msg.getEventType() == Notification.ADD) {
					notifyContainerAdded((Container) msg.getNewValue());
				} else if (msg.getEventType() == Notification.ADD_MANY) {
					List<?> newValues = (List<?>) msg.getNewValue();
					for (Object curValue : newValues) {
						notifyContainerAdded((Container) curValue);
					}
				} else if (msg.getEventType() == Notification.REMOVE) {
					notifyContainerRemoved((Container) msg.getOldValue());
				} else if (msg.getEventType() == Notification.REMOVE_MANY) {
					List<?> oldValues = (List<?>) msg.getOldValue();
					for (Object curValue : oldValues) {
						notifyContainerRemoved((Container) curValue);
					}
				}
			}
			// TODO: Also handle removals.
		}

		@Override
		public Agent getTarget() {
			return target;
		}

		@Override
		public void setTarget(Notifier newTarget) {
			target = (Agent) target;
		}

		@Override
		public boolean isAdapterForType(Object type) {
			return (type instanceof Agent);
		}
	}

	private Agent agent;
	private RepositoryConnection connection;
	private ModelRepository repository;
	private final ScheduledExecutorService workerThread = Executors.newScheduledThreadPool(1);
	private final String scopePath;
	private final InetAddress hostIP;
	private final int port;
	private ContainerScope scope;
	
	private volatile boolean running = false;

	public AgentController(String scopePath, InetAddress hostIP, int port) {
		this.scopePath = scopePath;
		this.hostIP = hostIP;
		this.port = port;
	}
	
	public Agent getAgent() {
		return agent;
	}

	public ContainerScope getContainerScope() {
		return scope;
	}

	public Future<?> start() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (!running) {
					startAsync();
				}
				running = true;
			}
		});
	}

	public Future<?> stop() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (running) {
					stopAsync();
				}
				running = false;
			}
		});
	}

	public void addContainerIncludeFilter(String property, String pattern) {
		checkRunning();
		try (Transaction transaction = getModelRepository().createTransaction()) {

			ContainerFilter filter = AgentFactory.eINSTANCE.createContainerFilter();
			filter.setProperty(property);
			filter.setFilter(pattern);
			Agent agentInTransaction = transaction.attach(agent);
			agentInTransaction.cdoWriteLock().lock();
			agentInTransaction.getIncludeFilters().add(filter);
			transaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
		scope.updateAgent(agent);
	}

	public void addContainerExcludeFilter(String property, String pattern) {
		checkRunning();
		try (Transaction transaction = getModelRepository().createTransaction()) {
			ContainerFilter filter = AgentFactory.eINSTANCE.createContainerFilter();
			filter.setProperty(property);
			filter.setFilter(pattern);
			Agent agentInTransaction = transaction.attach(agent);
			agentInTransaction.cdoWriteLock().lock();
			agentInTransaction.getExcludeFilters().add(filter);
			transaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
		scope.updateAgent(agent);
	}

	protected void startAsync() {
		connectToRepository();
		SystemGlobalScope systemScope = getModelRepository().getSystemScope();
		systemScope.create();
		this.scope = (ContainerScope) getModelRepository().getScope(scopePath);
		this.scope.create();
		this.agent = scope.addAgent(hostIP.getHostAddress(), port);
		this.agent.eAdapters().add(new ContainerListAdapter());
	}

	protected void stopAsync() {
		try {
			try (Transaction transaction = getModelRepository().createTransaction()) {
				AgentRepository repository = transaction.attach(scope.getAgents());
				repository.cdoWriteLock().lock();
				repository.getAgents().remove(agent);
				transaction.commit();
			} catch (CommitException e) {
				e.printStackTrace();
			}
		} finally {
			try {
				connection.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	protected ModelRepository getModelRepository() {
		return repository;
	}

	protected void connectToRepository() {
		this.connection = new RepositoryConnection("127.0.0.1");
		this.connection.open();
		this.repository = new CDOModelRepository(connection);
	}

	protected void notifyContainerAdded(Container newContainer) {
		System.out.println("New Container: " + newContainer.getName());
	}

	protected void notifyContainerRemoved(Container oldContainer) {
		System.out.println("Removed Container: " + oldContainer.getName());
	}

	protected void checkRunning() {
		if (!running) {
			throw new IllegalStateException("The agent is not running.");
		}
	}
}
