package tools.descartes.prisma.core.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.TimeoutException;

import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.cdo.common.lob.CDOClobWriter;
import org.eclipse.emf.cdo.eresource.CDOResourceFolder;
import org.eclipse.emf.cdo.eresource.CDOResourceNode;
import org.eclipse.emf.cdo.eresource.CDOTextResource;
import org.eclipse.emf.cdo.transaction.CDOTransaction;
import org.eclipse.emf.cdo.view.CDOAdapterPolicy;
import org.eclipse.emf.cdo.view.CDOView;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.prisma.core.ApplicationScope;
import tools.descartes.prisma.core.HypervisorScope;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.RepositoryConnection;
import tools.descartes.prisma.core.Scope;
import tools.descartes.prisma.core.SystemGlobalScope;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.exceptions.LockTimeoutException;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class CDOModelRepository implements ModelRepository {

	private final static long LOCK_TIMEOUT_IN_MILLIS = 60 * 1000;

	private final RepositoryConnection connection;
	private final CDOView readOnlyView;
	private final Map<String, Scope> scopes = new HashMap<>();
	private final SystemGlobalScope systemScope;

	public CDOModelRepository(RepositoryConnection connection) {
		this.connection = connection;
		this.readOnlyView = connection.startView();
		/*
		 * Important: To enable the notification of changes from other
		 * transactions/views change the subscription policy.
		 */
		this.readOnlyView.options().addChangeSubscriptionPolicy(CDOAdapterPolicy.CDO);
		this.systemScope = new SystemGlobalScope(this);
	}

	public SystemGlobalScope getSystemScope() {
		return systemScope;
	}

	public Scope getScope(String path) {
		Scope scope = scopes.get(path);
		if (scope == null) {
			scope = createScope(path);
			if (scope != null) {
				scopes.put(path, scope);
			}
		}
		return scope;
	}

	private Scope createScope(String path) {
		String[] parts = path.split("/");
		if (parts.length == 2) {
			if (parts[0].equals(ApplicationScope.APPLICATION_SCOPE_TYPE)) {
				return new ApplicationScope(parts[1], systemScope, this);
			} else if (parts[0].equals(HypervisorScope.HYPERVISOR_SCOPE_TYPE)) {
				return new HypervisorScope(parts[1], systemScope, this);
			}
		} else if (parts.length == 1) {
			if (parts[0].equals(systemScope.getName())) {
				return systemScope;
			}
		}
		return null;
	}

	@Override
	public void createFolders(Transaction transaction, String path) {
		CDOTransaction cdoTransaction = checkTransaction(transaction);

		String[] elements = path.split("/");
		if (elements.length < 1) {
			return;
		}
		CDOResourceNode[] nodes = cdoTransaction.getElements();
		CDOResourceFolder cur = null;
		for (CDOResourceNode curNode : nodes) {
			if (curNode.getName().equals(elements[0])) {
				cur = (CDOResourceFolder) curNode;
				break;
			}
		}
		if (cur == null) {
			acquireWriteLock(cdoTransaction.getRootResource());
			cur = cdoTransaction.createResourceFolder(elements[0]);
		}
		for (int i = 1; i < elements.length; i++) {
			CDOResourceFolder next = (CDOResourceFolder) cur.getNode(elements[i]);
			if (next == null) {
				acquireWriteLock(cur);
				next = cur.addResourceFolder(elements[i]);
			}
			cur = next;
		}
	}

	@Override
	public Resource getResource(String path) {
		return readOnlyView.getResource(path);
	}

	@Override
	public Resource getOrCreateResource(Transaction transaction, String path) {
		CDOTransaction cdoTransaction = checkTransaction(transaction);
		int idx = path.lastIndexOf("/");
		String basePath = "";
		String name = path;
		if (idx >= 0) {
			basePath = path.substring(0, idx);
			name = path.substring(idx + 1);
		}
		CDOResourceFolder folder = cdoTransaction.getResourceFolder(basePath);
		CDOResourceNode node = folder.getNode(name);
		if (node != null) {
			return (Resource) node;
		} else {
			acquireWriteLock(folder);
			return folder.addResource(name);
		}
	}

	@Override
	public List<String> listResources(String path) {
		CDOResourceFolder folder = readOnlyView.getResourceFolder(path);
		EList<CDOResourceNode> nodes = folder.getNodes();
		List<String> nodeNames = new ArrayList<String>(nodes.size());
		for (CDOResourceNode curNode : nodes) {
			nodeNames.add(curNode.getName());
		}
		return nodeNames;
	}

	@Override
	public boolean exists(String path) {
		CDOResourceFolder folder = null;
		if (!path.isEmpty()) {
			String[] elements = path.split("/");
			for (int i = 0; i < elements.length; i++) {
				List<CDOResourceNode> res = readOnlyView.queryResources(folder, elements[i], true);
				if (res.isEmpty()) {
					return false;
				}
				if (i == (elements.length - 1)) {
					return true;
				} else {
					CDOResourceNode node = res.get(0);
					if (node instanceof CDOResourceFolder) {
						folder = (CDOResourceFolder) node;
					} else {
						return false;
					}
				}
			}
		}
		return true;
	}

	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public Map<String, String> loadPropertiesFile(String path) throws IOException {
		Properties containerProperties = new Properties();
		CDOTextResource textFile = readOnlyView.getTextResource(path);
		containerProperties.load(textFile.getContents().getContents());
		// Awkward cast required here due to problems in Java API
		return (Map) containerProperties;
	}

	@Override
	public void storePropertiesFile(Transaction transaction, String path, Map<String, String> properties)
			throws IOException {
		CDOTransaction cdoTransaction = checkTransaction(transaction);

		// Save the properties in a text file for later retrieval
		CDOTextResource propRes = cdoTransaction.getOrCreateTextResource(path);
		acquireWriteLock(propRes);
		Properties prop = new Properties();
		prop.putAll(properties);
		CDOClobWriter writer = new CDOClobWriter();
		try {
			prop.store(writer, "");
		} finally {
			writer.close();
		}
		propRes.setContents(writer.getClob());
	}

	public Transaction createTransaction() {
		return new CDOTransactionWrapper(this, connection.startTransaction());
	}

	@Override
	public void waitForUpdate() {
		readOnlyView.waitForUpdate(readOnlyView.getLastUpdateTime() + 1);
	}

	public EObject adapt(EObject obj, ModelSkeleton skeleton) {
		Set<EObject> objectsToAdapt = new HashSet<EObject>();
		EObject root = obj;
		while (root != null) {
			objectsToAdapt.add(root);
			root = root.eContainer();
		}
		ModelSkeletonCopier copier = new ModelSkeletonCopier(objectsToAdapt);
		EObject rootCopy = copier.copy(root);
		copier.copyReferences();

		EcoreUtil.ExternalCrossReferencer.find(rootCopy);

		return null;
	}

	protected CDOView getReadOnlyView() {
		return readOnlyView;
	}

	private CDOTransaction checkTransaction(Transaction transaction) {
		if (transaction == null) {
			throw new IllegalArgumentException("Transaction may not be null.");
		}
		return ((CDOTransactionWrapper) transaction).getWrappedTransaction();
	}

	/**
	 * We do not use the standard {@link EcoreUtil#copy(EObject)} function since
	 * we do not do a complete copy. Only attributes and non-containment
	 * references are copied to the new object.
	 * 
	 * @param obj
	 * @return
	 */
	private EObject copy(EObject obj) {
		EObject copy = EcoreUtil.create(obj.eClass());
		for (EStructuralFeature feature : obj.eClass().getEStructuralFeatures()) {
			if (feature.isChangeable() && !feature.isDerived()) {
				if (feature instanceof EAttribute) {
					if (obj.eIsSet(feature)) {
						if (feature.isMany()) {
							List<?> source = (List<?>) obj.eGet(feature);
							@SuppressWarnings("unchecked")
							List<Object> target = (List<Object>) copy.eGet(feature);
							if (source.isEmpty()) {
								target.clear();
							} else {
								target.addAll(source);
							}
						} else {
							copy.eSet(feature, obj.eGet(feature));
						}
					}
				}
			}
		}
		return copy;
	}

	private void acquireWriteLock(CDOObject object) {
		try {
			object.cdoWriteLock().lock(LOCK_TIMEOUT_IN_MILLIS);
		} catch (TimeoutException e) {
			throw new LockTimeoutException(e);
		}
	}

}
