package tools.descartes.prisma.core.templates;

import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

public abstract class ModelTemplate<P extends EObject, E extends EObject> {
	
	public enum Status {
		CREATE, EXISTS, REMOVE
	}
	
	private final EStructuralFeature container;
	private final Class<E> elementType;
	private TemplateContext context;
	private final Status status;
	
	public ModelTemplate(EStructuralFeature container, Class<E> elementType) {
		this(Status.CREATE, container, elementType);
	}
	
	public ModelTemplate(Status status, EStructuralFeature container, Class<E> elementType) {
		this.container = container;
		this.elementType = elementType;
		this.status = status;
	}
	
	public void setContext(TemplateContext ctx) {
		this.context = ctx;
		for (ModelTemplate<?,?> cmd : getChildCommands()) {
			cmd.setContext(context);
		}
	}
	
	public TemplateContext getContext() {
		return context;
	}
	
	public abstract boolean matches(E object);
	
	protected E matchesWithCast(Object object) {
		if (elementType.isAssignableFrom(object.getClass())) {
			E casted = elementType.cast(object);
			if (matches(casted)) {
				return casted;
			}
		}
		return null;
	}
	
	public E find(P parent) {
		if (parent == null) {
			return null;
		}
		
		if (container.isMany()) {
			return find((List<?>)parent.eGet(container), 0);
		} else {
			Object value = parent.eGet(container);
			if (value != null) {
				E casted = matchesWithCast(value);
				if (casted != null) {
					return casted;
				}
			}
			return null;
		}		
	}
	
	protected E find(List<?> list) {
		return find(list, 0);
	}
	
	protected E find(List<?> list, int startIdx) {
		for (int i = startIdx; i < list.size(); i++) {
			E casted = matchesWithCast(list.get(i));
			if (casted != null) {
				return casted;
			}
		}
		return null;
	}
	
	protected void insert(P parent, E child) {
		if (container.isMany()) {
			// The EList implementation will check the type when inserting elements at runtime
			// this cast is therefore safe.
			@SuppressWarnings("unchecked")
			EList<E> list = ((EList<E>)parent.eGet(container));
			list.add(child);
		} else {
			parent.eSet(container, child);
		}
	}
	
	protected abstract E create(P parent);
	
	public List<? extends ModelTemplate<? super E, ?>> getChildCommands() {
		return Collections.emptyList();
	}
	
	public void apply(P parent) {
		E current = get(parent);
		
		if (current != null) {
			for(ModelTemplate<? super E, ?> cmd : getChildCommands()) {
				cmd.apply(current);
			}
		}
	}
	
	public E get(P parent) {
		E ret = find(parent);
		if ((status == Status.CREATE) && (ret == null)) {
			ret = create(parent);
			if (ret != null) {
				insert(parent, ret);
			}
		}
		return ret;
	}
	
	public Status getStatus() {
		return status;
	}
	
	protected Class<E> getElementType() {
		return elementType;
	}
	
	protected String getId(String name) {
		return name.replace("/", "_").replace(":", "_").replace(".",  "_").replace("#", "_").replace(" ", "_");
	}
	
	protected <T extends EObject> void apply(T parent, List<? extends ModelTemplate<T, ?>> childCommands) {
		for (ModelTemplate<T, ?> cmd : childCommands) {
			if (cmd.get(parent) == null) {
				cmd.create(parent);
			}
		}
	}	
}
